"""
Tests localbuild.py
"""

from subprocess import STDOUT
from unittest.mock import call

import pytest

from brewblox_tools import localbuild

TESTED = localbuild.__name__


@pytest.fixture
def mocked_utils(mocker):
    mocked = [
        'glob',
        'getenv',
        'remove',
        'check_call',
        'check_output',
        'find_dotenv',
        'load_dotenv',
    ]
    return {k: mocker.patch(TESTED + '.' + k) for k in mocked}


@pytest.fixture
def distcopy_mock(mocker):
    return mocker.patch(TESTED + '.distcopy.main')


@pytest.fixture
def run_mock(mocker):
    return mocker.patch(TESTED + '.run')


def test_run(mocked_utils):
    localbuild.run('test test one two')
    mocked_utils['check_call'].assert_called_once_with(
        'test test one two', shell=True, stderr=STDOUT
    )


def test_localbuild_simple(mocked_utils, distcopy_mock, run_mock):
    mocked_utils['glob'].return_value = ['f1', 'f2']
    mocked_utils['getenv'].side_effect = [
        'bb-repo',
    ]

    localbuild.main([])

    assert distcopy_mock.call_args_list == [
        call('dist/ docker/dist/'.split()),
        call('config/ docker/config/'.split()),
    ]
    assert mocked_utils['remove'].call_args_list == [
        call('f1'),
        call('f2'),
    ]
    assert run_mock.call_args_list == [
        call('python setup.py sdist'),
        call('pipenv lock --requirements > docker/requirements.txt'),
        call('docker build --no-cache --tag bb-repo:local --file docker/amd/Dockerfile docker')
    ]


def test_localbuild_all(mocked_utils, distcopy_mock, run_mock):
    mocked_utils['glob'].return_value = ['f1', 'f2']
    mocked_utils['getenv'].side_effect = [
        'bb-repo',
        'feature/funky_branch',
    ]

    localbuild.main(['--arch', 'amd', 'arm',
                     '--tags', 'test:tag',
                     '--push',
                     '--branch-tag',
                     '--pull',
                     '--context', 'dk',
                     '--file', 'df',
                     ])

    assert distcopy_mock.call_args_list == [
        call('dist/ dk/dist/'.split()),
        call('config/ dk/config/'.split()),
    ]
    assert run_mock.call_args_list == [
        call('python setup.py sdist'),
        call('pipenv lock --requirements > dk/requirements.txt'),
        call('docker build --pull --no-cache ' +
             '--tag bb-repo:local ' +
             '--tag bb-repo:test-tag ' +
             '--tag bb-repo:feature-funky-branch ' +
             '--file dk/amd/df ' +
             'dk'),
        call('docker push bb-repo:test-tag'),
        call('docker push bb-repo:feature-funky-branch'),
        call('docker run --rm --privileged multiarch/qemu-user-static:register --reset ' +
             '&& ' +
             'docker build --pull --no-cache ' +
             '--tag bb-repo:rpi-local ' +
             '--tag bb-repo:rpi-test-tag ' +
             '--tag bb-repo:rpi-feature-funky-branch ' +
             '--file dk/arm/df ' +
             'dk'),
        call('docker push bb-repo:rpi-test-tag'),
        call('docker push bb-repo:rpi-feature-funky-branch'),
    ]



"""
Tests dist_copy.py
"""

from unittest.mock import call

import pytest

from brewblox_tools import distcopy

TESTED = distcopy.__name__


@pytest.fixture
def mock_glob(mocker):
    return mocker.patch(TESTED + '.glob.glob')


@pytest.fixture
def mock_path(mocker):
    return mocker.patch(TESTED + '.pathlib.Path')


@pytest.fixture
def mock_copy(mocker):
    return mocker.patch(TESTED + '.shutil.copy')


@pytest.fixture
def mock_remove(mocker):
    return mocker.patch(TESTED + '.os.remove')


def test_distcopy(mock_glob, mock_path, mock_copy, mock_remove):
    mock_glob.return_value = ['f1', 'f2']

    distcopy.main('src dest1 dest2/'.split())

    mock_path.assert_has_calls([
        call('dest1'),
        call().mkdir(parents=True, exist_ok=True),
        call('dest2/'),
        call().mkdir(parents=True, exist_ok=True),
    ])

    mock_remove.assert_has_calls([
        call('f1'),
        call('f2')
    ])

    mock_copy.assert_has_calls([
        call('f1', 'dest1/'),
        call('f2', 'dest1/'),
        call('f1', 'dest2//'),
        call('f2', 'dest2//')
    ])

    mock_glob.assert_has_calls([
        call('dest1/*'),
        call('src/*'),
        call('dest2//*'),
        call('src/*')
    ])



#! /usr/bin/python3

import argparse
from distutils.util import strtobool
from distutils.version import StrictVersion
from subprocess import STDOUT, check_call, check_output


def parse_args(sys_args: list = None):
    argparser = argparse.ArgumentParser()
    argparser.add_argument('bump_type',
                           help='Component of the version number to increment.',
                           choices=['major', 'minor', 'patch'])
    return argparser.parse_args(sys_args)


def user_yes_no_query(question: str) -> bool:
    print(f'{question} [Y/n]')
    while True:
        try:
            return strtobool(input().lower() or 'yes')
        except ValueError:
            print('Please respond with \'y\' or \'n\'.')


def bump(current_version: str, bump_type: str) -> str:
    major, minor, patch = StrictVersion(current_version).version

    return {
        'major': lambda: f'{major + 1}.{0}.{0}',
        'minor': lambda: f'{major}.{minor + 1}.{0}',
        'patch': lambda: f'{major}.{minor}.{patch + 1}',
    }[bump_type]()


def main(sys_args: list = None):
    args = parse_args(sys_args)
    print(vars(args))

    # Get all version-formatted tags, but use the latest
    current_version = check_output(
        r'git tag -l *.*.* --contains $(git rev-list --tags --max-count=1)',
        shell=True
    ).decode().rstrip().split('\n')[-1]

    new_version = bump(current_version, args.bump_type)

    print(f'Bumping "{args.bump_type}" version: {current_version} ==> {new_version}')

    if user_yes_no_query('Do you want to tag the current commit with that version?'):
        check_output(f'git tag -a {new_version} -m "Version {new_version}"', shell=True)

        print('Latest tags:')
        print(check_output('git tag -n1 | tail -n5', shell=True).decode().rstrip())

    else:
        print('Aborted. No tags were added!')
        return

    if user_yes_no_query('Do you want to push this tag?'):
        check_call('git push --tags', shell=True, stderr=STDOUT)


if __name__ == '__main__':
    main()



#! /usr/bin/python3

import argparse
import json
import re

import docker

TEMP_TAG = 'temp'


def parse_args(sys_args: list = None):
    argparser = argparse.ArgumentParser()
    argparser.add_argument('-f', '--file', help='Dockerfile to be built. [%(default)s]', default='Dockerfile')
    argparser.add_argument('-c', '--context', help='Docker build context directory. [%(default)s]', default='.')
    argparser.add_argument('-i', '--image', help='DEPRECATED: use --context instead', default='.', dest='context')
    argparser.add_argument('-n', '--name', help='Image base name', required=True)
    argparser.add_argument('-t', '--tags', nargs='+', help='Tags to add to built image.', default=[])
    argparser.add_argument('--no-cache', help='Don\'t use cache when building.', action='store_true')
    argparser.add_argument('--no-push', help='Don\'t push added tags to Docker Hub.', action='store_true')
    return argparser.parse_args(sys_args)


def build(client, args):
    print('\n==== BUILDING ====\n')

    # We're using the low-level API client to get a continuous stream of messages
    tag_name = f'{args.name}:{TEMP_TAG}'

    generator = client.build(
        path=args.context,
        dockerfile=args.file,
        tag=tag_name,
        rm=True,
        nocache=args.no_cache)

    while True:
        try:
            output = next(generator).rstrip()
            json_output = json.loads(output)
            if 'stream' in json_output:
                print(json_output['stream'].rstrip(), flush=True)
        except StopIteration:
            print('Docker image build complete.')
            break
        except ValueError:
            print(f'Error parsing output from docker image build: {output}')


def deploy(client, args):
    if args.tags:
        print('\n==== DEPLOYING ====\n')

    for tag in args.tags:
        # Filter out illegal tag characters
        tag = re.sub('[/_:]', '-', tag)
        print(f'Tagging {args.name}:{tag}')
        client.tag(f'{args.name}:{TEMP_TAG}', repository=args.name, tag=tag)

        if args.no_push:
            continue

        print(f'Pushing {args.name}:{tag}')
        for line in client.push(repository=args.name, tag=tag, stream=True):
            output = line.rstrip()
            json_output = json.loads(output)
            msg = ' '.join([str(v) for v in json_output.values()])
            print(msg, flush=True)


def main(sys_args: list = None):
    args = parse_args(sys_args)
    print('+'*20, 'Deploy Docker Image', '+'*20)
    print(args)
    docker_client = docker.APIClient()

    build(docker_client, args)
    deploy(docker_client, args)


if __name__ == '__main__':
    main()



#! /usr/bin/python3

import argparse
import glob
import os
import pathlib
import shutil


def parse_args(sys_args: list = None):
    argparser = argparse.ArgumentParser()
    argparser.add_argument('source', help='Source directory')
    argparser.add_argument('destinations', nargs='+', help='Destination directories.')
    return argparser.parse_args(sys_args)


def main(sys_args: list = None):
    args = parse_args(sys_args)
    print(vars(args))

    for dest in args.destinations:
        pathlib.Path(dest).mkdir(parents=True, exist_ok=True)

        for f in glob.glob(dest + '/*'):
            os.remove(f)

        for f in glob.glob(args.source + '/*'):
            shutil.copy(f, dest + '/')
            print(f'{f} ==> {dest}')


if __name__ == '__main__':
    main()



"""
Tests bump.py
"""

from subprocess import STDOUT

import pytest

from brewblox_tools import bump

TESTED = bump.__name__


@pytest.fixture
def check_output_mock(mocker):
    m = mocker.patch(TESTED + '.check_output')
    return m


@pytest.fixture
def check_call_mock(mocker):
    m = mocker.patch(TESTED + '.check_call')
    return m


@pytest.fixture
def input_mock(mocker):
    m = mocker.patch('builtins.input')
    return m


@pytest.mark.parametrize('bump_type,new_version', [
    ('major', '4.0.0'),
    ('minor', '3.3.0'),
    ('patch', '3.2.2')
])
def test_bump(check_output_mock, check_call_mock, input_mock, bump_type, new_version):
    check_output_mock.return_value = b'1.2.3\n3.2.1\n'
    input_mock.return_value = 'y'

    bump.main([bump_type])

    check_output_mock.assert_any_call(f'git tag -a {new_version} -m "Version {new_version}"', shell=True)
    check_call_mock.assert_any_call('git push --tags', shell=True, stderr=STDOUT)
    assert input_mock.call_count == 2


def test_bump_no_push(check_output_mock, input_mock, check_call_mock):
    check_output_mock.return_value = b'1.2.3\n3.2.1\n'
    input_mock.side_effect = ['y', 'n']

    bump.main(['minor'])

    check_output_mock.assert_any_call(f'git tag -a 3.3.0 -m "Version 3.3.0"', shell=True)
    assert input_mock.call_count == 2
    assert check_call_mock.call_count == 0


def test_bump_nok(check_output_mock, input_mock):
    check_output_mock.return_value = b'1.2.3\n3.2.1\n'
    input_mock.return_value = 'n'

    bump.main(['minor'])
    assert check_output_mock.call_count == 1


@pytest.mark.parametrize('user_input,expected', [
    (['y'], True),
    (['n'], False),
    (['Y'], True),
    (['N'], False),
    (['Yes'], True),
    (['No'], False),
    (['y', 'n'], True),
    (['n', 'y'], False),
    (['maybe', 'y'], True),
    (['maybe', 'I think so...', 'n'], False),
    ([''], True),
])
def test_user_yes_no_query(input_mock, user_input, expected):
    input_mock.side_effect = user_input
    assert bump.user_yes_no_query('Are you sure?') == expected



#! /usr/bin/python3


import argparse
import re
from glob import glob
from os import getenv, remove
from subprocess import STDOUT, check_call, check_output

from dotenv import find_dotenv, load_dotenv

from brewblox_tools import distcopy

# Import various OS libraries as special name, to allow mocking them in unit tests
# Otherwise, pytest will break as it starts using mocked functions


def parse_args(sys_args: list = None):
    parser = argparse.ArgumentParser()

    parser.add_argument('-a', '--arch',
                        help='Build these architectures. [%(default)s]',
                        default=['amd'],
                        nargs='+',
                        choices=['amd', 'arm'])
    parser.add_argument('-r', '--repo',
                        help='Docker repository name. [%(default)s]',
                        default=getenv('DOCKER_REPO'))
    parser.add_argument('-c', '--context',
                        help='Build context. [%(default)s]',
                        default='docker')
    parser.add_argument('-f', '--file',
                        help='Filename inside context. [%(default)s]',
                        default='Dockerfile')
    parser.add_argument('-t', '--tags',
                        help='Additional tags. The "local" tag is always built.',
                        nargs='+',
                        default=[])
    parser.add_argument('--branch-tag',
                        help='Use sanitized branch name as tag. ' +
                        'ARM automatically gets the "rpi-" prefix. [%(default)s]',
                        action='store_true')
    parser.add_argument('--pull',
                        help='Pull base images. [%(default)s]',
                        action='store_true')
    parser.add_argument('--push',
                        help='Push all tags except "local" to Docker Hub. [%(default)s]',
                        action='store_true')

    return parser.parse_args(sys_args)


def run(cmd: str):
    print(cmd)
    check_call(cmd, shell=True, stderr=STDOUT)


def main(sys_args: list = None):
    load_dotenv(find_dotenv(usecwd=True))
    args = parse_args(sys_args)
    tags = args.tags.copy()

    if args.branch_tag:
        tags.append(getenv('TRAVIS_BRANCH')
                    or check_output('git rev-parse --abbrev-ref HEAD'.split()).decode().rstrip())

    tags = [re.sub('[/_:]', '-', tag) for tag in tags]

    for f in glob('dist/*'):
        remove(f)

    run('python setup.py sdist')
    run(f'pipenv lock --requirements > {args.context}/requirements.txt')
    distcopy.main(f'dist/ {args.context}/dist/'.split())
    distcopy.main(f'config/ {args.context}/config/'.split())

    for arch in args.arch:
        prefix = ''
        commands = []
        build_args = []
        build_tags = [
            'local',
            *tags
        ]

        if arch == 'arm':
            prefix = 'rpi-'
            commands += [
                'docker run --rm --privileged multiarch/qemu-user-static:register --reset',
            ]

        if args.pull:
            build_args.append('--pull')

        build_args += [
            '--no-cache',
            ' '.join([f'--tag {args.repo}:{prefix}{t}' for t in build_tags]),
            f'--file {args.context}/{arch}/{args.file}',
            args.context,
        ]

        commands.append(f'docker build {" ".join(build_args)}')
        run(' && '.join(commands))

        if args.push:
            # We're skipping the local tag
            [run(f'docker push {args.repo}:{prefix}{t}') for t in tags]


if __name__ == '__main__':
    main()



"""
Tests deploy_docker.py
"""

from unittest.mock import call

import pytest

from brewblox_tools import deploy_docker

TESTED = deploy_docker.__name__


BUILD_OUTPUT = [
    b'{"stream":"Step 1/6 : FROM python:3.6"}',
    b'{"stream":"\n"}',
    b'{"stream":" ---\\u003e d21927554614\n"}',
    b'{"stream":"Step 2/6 : RUN mkdir -p /pkg"}',
    b'{"stream":"\n"}',
    b'{"something":"text"}',
    b'not_json',
    b'{"stream":" ---\\u003e Running in 7b0a59f4706b\n"}',
    b'{"stream":"Removing intermediate container 7b0a59f4706b\n"}',
    b'{"stream":" ---\\u003e 402bcb59fe89\n"}',
    b'{"stream":"Step 3/6 : COPY ./pkg/* /pkg/"}',
    b'{"stream":"\n"}',
    b'{"stream":" ---\\u003e 31c862396aca\n"}',
    b'{"stream":"Step 4/6 : EXPOSE 5000"}',
    b'{"stream":"\n"}',
    b'{"stream":" ---\\u003e Running in 7d6552578599\n"}',
    b'{"stream":"Removing intermediate container 7d6552578599\n"}',
    b'{"stream":" ---\\u003e a2b7dc3f9f5d\n"}',
]

PUSH_OUTPUT = [
    b'{"status":"The push refers to repository [docker.io/brewblox/brewblox-service]"}',
    b'{"status":"Preparing","progressDetail":{},"id":"64d234a0ce4c"}',
    b'{"status":"Preparing","progressDetail":{},"id":"2ca1d28f3a22"}',
    b'{"status":"Preparing","progressDetail":{},"id":"e4b347e25e5d"}',
    b'{"status":"Preparing","progressDetail":{},"id":"aec4f1507d85"}',
    b'{"status":"Pushing","progressDetail":{"current":101888,"total":98321},"progress":"[=============='
    b'====================================\\u003e]  101.9kB","id":"2ca1d28f3a22"}',
    b'{"status":"Pushing","progressDetail":{"current":267776,"total":25576721},"progress":"[\\u003e      '
    b'                                            ]  267.8kB/25.58MB","id":"64d234a0ce4c"}',
    b'{"status":"Pushing","progressDetail":{"current":800578,"total":25576721},"progress":"[=\\u003e      '
    b'                                           ]  800.6kB/25.58MB","id":"64d234a0ce4c"}',
    b'{"status":"Layer already exists","progressDetail":{},"id":"6e1b48dc2ccc"}',
    b'{"status":"Pushing","progressDetail":{"current":1593856,"total":25576721},"progress":"[===\\u003e    '
    b'                                           ]  1.594MB/25.58MB","id":"64d234a0ce4c"}',
]


@pytest.fixture
def client_mock(mocker):
    m = mocker.patch(TESTED + '.docker.APIClient').return_value
    m.build.return_value = iter(BUILD_OUTPUT)
    m.push.return_value = iter(PUSH_OUTPUT)
    return m


def test_build(client_mock):
    deploy_docker.main('-n repo-name -i dir-name'.split())
    client_mock.build.assert_called_once_with(
        path='dir-name',
        dockerfile='Dockerfile',
        tag='repo-name:temp',
        rm=True,
        nocache=False
    )


def test_deploy(client_mock):
    deploy_docker.main('-n repo-name -t humpty dumpty'.split())
    client_mock.tag.assert_has_calls([
        call('repo-name:temp', repository='repo-name', tag='humpty'),
        call('repo-name:temp', repository='repo-name', tag='dumpty'),
    ])

    client_mock.push.assert_has_calls([
        call(repository='repo-name', tag='humpty', stream=True),
        call(repository='repo-name', tag='dumpty', stream=True),
    ])


def test_no_push(client_mock):
    deploy_docker.main('-n repo-name -t humpty dumpty --no-push'.split())
    client_mock.tag.assert_has_calls([
        call('repo-name:temp', repository='repo-name', tag='humpty'),
        call('repo-name:temp', repository='repo-name', tag='dumpty'),
    ])

    assert client_mock.push.call_count == 0
